#include <ctype.h>
#include <getopt.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parsing.h"

void free_options(struct option_lst *options)
{
	struct option_lst *it = options;
	while (it) {
		struct option_lst *tmp = it;
		it = it->next;
		free(tmp);
	}
}

bool option_isset(const struct option_lst *options, e_flag flag)
{
	if (!options)
		return false;
	for (const struct option_lst *it = options; it; it = it->next)
		if (it->flag == flag)
			return true;
	return false;
}

char *get_option_arg(const struct option_lst *options, e_flag flag)
{
	if (!options)
		return NULL;
	for (const struct option_lst *it = options; it; it = it->next)
		if (it->flag == flag)
			return it->arg;
	return NULL;
}

e_scantype parse_type(const char *arg)
{
	const char *types[] = {
	    "NULL", "SYN", "ACK", "FIN", "XMAS", "UDP",
	};

	if (!arg)
		return SCAN_ALL;
	for (size_t i = 0; i < 6; i++)
		if (!strcmp(arg, types[i]))
			return i;
	return -1;
}

int parse_ports(const char *arg, uint16_t *p_start, uint16_t *p_end)
{
	size_t i = 0;
	if (!arg)
		return -1;
	for (; isdigit(arg[i]); i++)
		;
	int start = atoi(arg);
	if (arg[i] == '\0') {
		*p_start = start;
		*p_end = start;
		return 0;
	}
	if (arg[i++] != '-')
		goto error;
	int end = atoi(arg + i);
	for (; isdigit(arg[i]); i++)
		;
	if (arg[i])
		goto error;
	if (start >= end)
		goto error;
	if (end > UINT16_MAX) {
		dprintf(2,
		        "ft_nmap: invalid argument: '%s': out of range: 0 "
		        "<= value <= %d\n",
		        arg, UINT16_MAX);
		return -1;
	}
	*p_start = start;
	*p_end = end;
	return 0;
error:
	dprintf(2, "ft_nmap: invalid argument: '%s'\n", arg);
	return -1;
}

static int check_arg(e_flag flag, const char *arg)
{
	switch (flag) {
	case FL_SPEEDUP: {
		for (size_t i = 0; arg[i]; i++)
			if (!isdigit(arg[i]))
				return -1;
		int value = atoi(arg);
		if (value < 0 || value > 250) {
			dprintf(2,
			        "ft_nmap: invalid argument: '%s': out "
			        "of range: 0 "
			        "<= value <= 250\n",
			        arg);
			return -1;
		}
		break;
	}
	case FL_SCAN:
		if (!strcmp(arg, "SYN"))
			break;
		if (!strcmp(arg, "NULL"))
			break;
		if (!strcmp(arg, "ACK"))
			break;
		if (!strcmp(arg, "FIN"))
			break;
		if (!strcmp(arg, "XMAS"))
			break;
		if (!strcmp(arg, "UDP"))
			break;
		return -1;
	case FL_TTL: {
		int nb = atoi(arg);
		if (nb == 0) {
			dprintf(2, "ft_nmap: cannot set unicast time-to-live: "
			           "Invalid argument\n");
			return -1;
		} else if (nb < 0 || nb > 255) {
			dprintf(
			    2,
			    "ft_nmap: invalid argument: '%s': out of range: 0 "
			    "<= value <= 255\n",
			    arg);
			return -1;
		}
		break;
	}
	case FL_MAXRETRIES: {
		int nb = atoi(arg);
		if (nb == 0) {
			dprintf(2, "ft_nmap: cannot set max_retries to 0\n");
			return -1;
		} else if (nb < 0 || nb > 255) {
			dprintf(
			    2,
			    "ft_nmap: invalid argument: '%s': out of range: 0 "
			    "<= value <= 255\n",
			    arg);
			return -1;
		}
		break;
	}
	default:
		break;
	}
	return 0;
}

static int add_option(struct option_lst **head, e_flag flag, char *arg)
{
	struct option_lst *new_option = malloc(sizeof(struct option_lst));
	if (!new_option) {
		dprintf(2, "ft_nmap: allocation of option failed\n");
		for (struct option_lst *it = *head; it;) {
			struct option_lst *tmp = it;
			it = it->next;
			free(tmp);
		}
		return -1;
	}

	new_option->flag = flag;
	new_option->arg = arg;
	new_option->next = NULL;

	if (*head == NULL) {
		*head = new_option;
		return 0;
	}
	for (struct option_lst *it = *head; it; it = it->next) {
		if (it->next == NULL) {
			it->next = new_option;
			break;
		}
	}
	return 0;
}

struct option_lst *parse_options(int ac, char *const *av)
{
	struct option_lst *head = NULL;
	const struct option options[] = {
	    {"help", no_argument, 0, 0},
	    {"ports", required_argument, 0, 0},
	    {"ip", required_argument, 0, 0},
	    {"file", required_argument, 0, 0},
	    {"speedup", required_argument, 0, 0},
	    {"scan", required_argument, 0, 0},
	    {"max_retries", required_argument, 0, 0},
	    {"ttl", required_argument, 0, 0},
	};

	int c;
	while (1) {
		int option_index = 0;
		c = getopt_long(ac, av, "F", options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'F':
			add_option(&head, FL_FAST, optarg);
			break;
		case '?':
			return NULL;
		default:
			break;
		}
		if (check_arg(option_index, optarg) < 0)
			return NULL;
		if (add_option(&head, option_index, optarg) < 0)
			return NULL;
	}
	return head;
}
